@@ -39,6 +39,9 @@
 #include <sys/signal.h>
 #include <fcntl.h>
 
+#define D(x)
+/* #define D(x) g_print x */
+
 #define LINE_LENGTH 4096	/* max line length we'll grok */
 #define SLEEP_TIME 300000	/* time we'll wait for the process to finish */
 
@@ -52,6 +55,7 @@ typedef struct {
 	GnomeVFSOpenMode open_mode;
 	int read_fd;
 	int write_fd;
+	int error_fd;
 	pid_t pid;
 	GnomeVFSFileInfoOptions info_opts;
 } SshHandle;
@@ -111,16 +115,19 @@ static GnomeVFSResult do_make_directory (GnomeVFSMethod *method,
 static GnomeVFSResult do_remove_directory(GnomeVFSMethod *method,
 					  GnomeVFSURI *uri,
 					  GnomeVFSContext *context);
+static GnomeVFSResult do_move            (GnomeVFSMethod *method,
+					  GnomeVFSURI *old_uri,
+					  GnomeVFSURI *new_uri,
+					  gboolean force_replace,
+					  GnomeVFSContext *context);
 static GnomeVFSResult do_unlink         (GnomeVFSMethod *method,
 					 GnomeVFSURI *uri,
 					 GnomeVFSContext *context);
-#if 0
 static GnomeVFSResult do_check_same_fs	(GnomeVFSMethod  *method,
 					 GnomeVFSURI     *a,
 					 GnomeVFSURI     *b,
 					 gboolean        *same_fs_return,
 					 GnomeVFSContext *context);
-#endif
 static GnomeVFSResult do_set_file_info  (GnomeVFSMethod *method,
 					 GnomeVFSURI *uri,
 					 const GnomeVFSFileInfo *info,
@@ -153,9 +160,9 @@ static GnomeVFSMethod method = {
         do_is_local,
 	do_make_directory, /* make directory */
         do_remove_directory, /* remove directory */
-	NULL, /* move */
+	do_move,
 	do_unlink, /* unlink */
-	NULL, /* do_check_same_fs */
+	do_check_same_fs,
 	do_set_file_info, /* set_file_info */
 	NULL, /* truncate */
 	NULL, /* find_directory */
@@ -171,10 +178,11 @@ ssh_connect (SshHandle **handle_return,
 	SshHandle *handle;
 	char *command_line, *host_port;
 	const gchar *username, *hostname;
-	int argc, error_fd;
+	int argc;
 	GError *gerror = NULL;
 	GnomeVFSFileSize bytes_read;
 	char buffer[LINE_LENGTH];
+	GnomeVFSResult res;
 
 	/* We do support ssh:/// as ssh://localhost/ */
 	hostname = gnome_vfs_uri_get_host_name (uri);
@@ -194,10 +202,14 @@ ssh_connect (SshHandle **handle_return,
 				     username,
 				     " -p ", host_port,
 				     " ", hostname,
-				     " ", "\"LC_ALL=C;", command,"\"",
+				     " ",
+				     "\"LC_ALL=C; echo READY > /dev/stderr;",
+				     command, "; echo DONE > /dev/stderr\"",
 				     NULL);
 	g_free (host_port);
 
+	D(("ssh_connect () command: %s\n", command_line));
+
 	g_shell_parse_argv (command_line, &argc, &argv, &gerror);
 
 	g_free (command_line);
@@ -212,8 +224,9 @@ ssh_connect (SshHandle **handle_return,
 	g_spawn_async_with_pipes (NULL, argv, NULL, 
 				  G_SPAWN_SEARCH_PATH,
 				  NULL, NULL,
-				  &handle->pid, &handle->write_fd, &handle->read_fd,
-				  &error_fd, &gerror);
+				  (gint *)&handle->pid, &handle->write_fd, 
+				  &handle->read_fd, &handle->error_fd, 
+				  &gerror);
 	g_strfreev (argv);
 
 	if (gerror) {
@@ -228,24 +241,25 @@ ssh_connect (SshHandle **handle_return,
 
 	/* You can add more error checking here */
 	memset (buffer, '\0', LINE_LENGTH);
-	ssh_read_error (error_fd, &buffer, LINE_LENGTH, &bytes_read);
+	res = ssh_read_error (handle->error_fd, &buffer,
+			LINE_LENGTH, &bytes_read);
 
-	if (bytes_read != 0) {
-		if (strncmp ("Permission denied", buffer,
+	if (bytes_read != 0 && res == GNOME_VFS_OK) {
+		if (strncmp ("READY", buffer, strlen ("READY")) == 0) {
+			res = GNOME_VFS_OK;
+		} else if (strncmp ("Permission denied", buffer,
 					strlen("Permission denied")) == 0) {
-			close (error_fd);
-			return GNOME_VFS_ERROR_LOGIN_FAILED;
-		}
-		if (strncmp ("Host key verification failed", buffer,
-					strlen("Host key verification failed")) == 0) {
-			close (error_fd);
-			return GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
+			res = GNOME_VFS_ERROR_LOGIN_FAILED;
+		} else if (strncmp ("Host key verification failed", buffer,
+					strlen("Host key verification failed"))
+				== 0) {
+			res = GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
+		} else if (strstr (buffer, "Connection refused") != NULL) {
+			res = GNOME_VFS_ERROR_ACCESS_DENIED;
 		}
 	}
 
-	close (error_fd);
-
-	return GNOME_VFS_OK;
+	return res;
 }
 
 static GnomeVFSResult
@@ -253,6 +267,7 @@ ssh_destroy (SshHandle *handle)
 {
 	close (handle->read_fd);
 	close (handle->write_fd);
+	close (handle->error_fd);
 	gnome_vfs_uri_unref (handle->uri);
 	kill (handle->pid, SIGINT);
 	g_free (handle);
@@ -300,6 +315,26 @@ ssh_read_error (int error_fd,
 	return GNOME_VFS_OK;
 }
 
+static GnomeVFSResult
+ssh_check_for_done (SshHandle *handle)
+{
+	char buffer[LINE_LENGTH];
+	GnomeVFSResult res;
+	GnomeVFSFileSize bytes_read;
+
+	memset (buffer, '\0', LINE_LENGTH);
+	res = ssh_read_error (handle->error_fd, &buffer,
+			LINE_LENGTH, &bytes_read);
+
+	if (bytes_read != 0 && res == GNOME_VFS_OK) {
+		if (strncmp ("DONE", buffer, strlen ("DONE")) == 0) {
+			return res;
+		}
+	}
+
+	return GNOME_VFS_ERROR_GENERIC;
+}
+
 static GnomeVFSResult
 ssh_write (SshHandle *handle,
 	   gconstpointer buffer,
@@ -381,6 +416,9 @@ do_open (GnomeVFSMethod *method,
 		char *quoted_name;
 		name = gnome_vfs_unescape_string (uri->text, 
 						  G_DIR_SEPARATOR_S);
+		if (name == NULL) {
+			return GNOME_VFS_ERROR_INVALID_URI;
+		}
 		quoted_name = g_shell_quote (name);
 		g_free (name);
 
@@ -422,6 +460,9 @@ do_create (GnomeVFSMethod *method,
 	}
 
 	name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
+	if (name == NULL) {
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
 	quoted_name = g_shell_quote (name);
 
 	cmd = g_strdup_printf ("cat > %s", quoted_name);
@@ -459,7 +500,7 @@ do_read (GnomeVFSMethod *method,
 	 GnomeVFSFileSize *bytes_read,
 	 GnomeVFSContext *context)
 {
-	GnomeVFSResult result;
+	GnomeVFSResult result, res_secondary;
 
 	result =  ssh_read ((SshHandle *)method_handle, buffer, num_bytes,
 			bytes_read);
@@ -468,7 +509,12 @@ do_read (GnomeVFSMethod *method,
 	}
 
 	if (*bytes_read == 0) {	
-		result = GNOME_VFS_ERROR_EOF;
+		res_secondary = ssh_check_for_done ((SshHandle *)method_handle);
+		if (res_secondary != GNOME_VFS_OK) {
+			result = res_secondary;
+		} else {
+			result = GNOME_VFS_ERROR_EOF;
+		}
 	}
 	return result;
 }
@@ -502,6 +548,10 @@ do_open_directory (GnomeVFSMethod *method,
 	char *quoted_name;
 
 	name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
+	if (name == NULL) {
+		return GNOME_VFS_ERROR_NOT_FOUND;
+	}
+
 	quoted_name = g_shell_quote (name);
 
 	if (*name != '\0') {
@@ -548,7 +598,9 @@ get_access_info (GnomeVFSURI *uri, GnomeVFSFileInfo *file_info)
                                {'w', GNOME_VFS_PERM_ACCESS_WRITABLE}};
 
      name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
-
+     if (name == NULL) {
+	     return;
+     }
 
      if ( *name == '\0' ) {
              quoted_name = g_shell_quote ("/");
@@ -596,13 +648,18 @@ do_read_directory (GnomeVFSMethod *method,
                    GnomeVFSFileInfo *file_info,
 		   GnomeVFSContext *context)
 {
-	GnomeVFSResult result = GNOME_VFS_OK;
+	GnomeVFSResult result = GNOME_VFS_OK, res_secondary;
 	char line[LINE_LENGTH+1];
 	char c;
 	int i=0;
 	GnomeVFSFileSize bytes_read;
 	struct stat st;
 	char *tempfilename, *filename, *linkname;
+	SshHandle *handle = (SshHandle *)method_handle;
+	const char *mime_type;
+	char *target_name, *parent;
+	GnomeVFSFileInfo *target_info;
+	GnomeVFSURI *target_uri;
 
 	for (;;) {
 		tempfilename = NULL;
@@ -612,14 +669,19 @@ do_read_directory (GnomeVFSMethod *method,
 		bytes_read = 0;
 
 		while (i<LINE_LENGTH) {
-			result = ssh_read ((SshHandle *)method_handle, &c,
+			result = ssh_read (handle, &c,
 					   sizeof(char), &bytes_read);
 			if (bytes_read == 0 || c == '\r' || c == '\n') {
 				break;
 			}
 
 			if (result != GNOME_VFS_OK) {
-				return result;
+				res_secondary = ssh_check_for_done (handle);
+				if (res_secondary != GNOME_VFS_OK) {
+					result = res_secondary;
+				} else {
+					return result;
+				}
 			}
 
 			line[i] = c;
@@ -651,17 +713,50 @@ do_read_directory (GnomeVFSMethod *method,
 		gnome_vfs_stat_to_file_info (file_info, &st);
 		file_info->name = filename;
 		if (linkname) {
+			GNOME_VFS_FILE_INFO_SET_SYMLINK (file_info, TRUE);
+			file_info->valid_fields
+				|= GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME;
 			file_info->symlink_name = linkname;
+			D(("symlink: %s\n", file_info->symlink_name));
 		}
 
-		/* FIXME: support symlinks correctly */
+		/* FIXME: support symlinks to directories correctly */
 
-		if (((SshHandle*)method_handle)->info_opts
-		    & GNOME_VFS_FILE_INFO_GET_MIME_TYPE) {
-			file_info->mime_type = g_strdup 
-				(gnome_vfs_get_file_mime_type (filename, &st, 
-							       FALSE));
-			file_info->valid_fields 
+		if (file_info->type == GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK) {
+			parent = gnome_vfs_uri_to_string
+				(handle->uri, 0);
+			target_name =
+				gnome_vfs_make_uri_full_from_relative
+				(parent, file_info->symlink_name);
+
+			if ((handle->info_opts
+				& GNOME_VFS_FILE_INFO_FOLLOW_LINKS) == 0) {
+				mime_type = "x-special/symlink";
+			} else {
+				target_uri = gnome_vfs_uri_new (target_name);
+				mime_type = gnome_vfs_get_file_mime_type
+					(target_name, NULL, FALSE);
+				D(("Looking up mime-type for '%s'\n\n",
+							target_name));
+				target_info = gnome_vfs_file_info_new ();
+				do_get_file_info (method, target_uri,
+						target_info, handle->info_opts,
+						context);
+				file_info->type = target_info->type;
+				gnome_vfs_file_info_unref (target_info);
+				gnome_vfs_uri_unref (target_uri);
+			}
+
+			g_free (target_name);
+			g_free (parent);
+		} else {
+			mime_type = (gnome_vfs_get_file_mime_type
+					(filename, &st, FALSE));
+		}
+
+		if (handle->info_opts & GNOME_VFS_FILE_INFO_GET_MIME_TYPE) {
+			file_info->mime_type = g_strdup (mime_type);
+			file_info->valid_fields
 				|= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
 		}
 
@@ -669,10 +764,8 @@ do_read_directory (GnomeVFSMethod *method,
 			~GNOME_VFS_FILE_INFO_FIELDS_BLOCK_COUNT;
 		file_info->valid_fields &= 
 			~GNOME_VFS_FILE_INFO_FIELDS_IO_BLOCK_SIZE;
-		if (((SshHandle*)method_handle)->info_opts 
-		    & GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS) {
-			get_access_info (((SshHandle*)method_handle)->uri, 
-					 file_info);
+		if (handle->info_opts & GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS) {
+			get_access_info (handle->uri, file_info);
 		}
 
 		/* Break out.
@@ -699,6 +792,10 @@ do_get_file_info (GnomeVFSMethod *method,
 	char *quoted_name;
 
 	name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
+	if (name == NULL) {
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
+		
 	quoted_name = g_shell_quote (name);
 
 	if (*name != '\0') {
@@ -740,6 +837,9 @@ do_make_directory (GnomeVFSMethod *method,
 	char *quoted_name;
 
 	name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
+	if (name == NULL) {
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
 	quoted_name = g_shell_quote (name);
 
 	cmd = g_strdup_printf ("mkdir %s", quoted_name);
@@ -769,8 +869,9 @@ do_remove_directory (GnomeVFSMethod *method,
 	gchar *quoted_name;
 
 	name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
-	if (name == NULL)
+	if (name == NULL) {
 		return GNOME_VFS_ERROR_INVALID_URI;
+	}
 	quoted_name = g_shell_quote (name);
 
 	cmd = g_strdup_printf ("rm -rf %s", quoted_name);
@@ -788,20 +889,68 @@ do_remove_directory (GnomeVFSMethod *method,
 	return result;
 }
 
+static GnomeVFSResult
+do_move (GnomeVFSMethod *method,
+	 GnomeVFSURI *old_uri,
+	 GnomeVFSURI *new_uri,
+	 gboolean force_replace,
+	 GnomeVFSContext *context)
+{
+	SshHandle *handle = NULL;
+	GnomeVFSResult res;
+	gboolean same_fs;
+	char *old_name, *new_name;
+	char *quoted_old_name, *quoted_new_name;
+	char *cmd;
+
+	res = do_check_same_fs (method, old_uri, new_uri, &same_fs, context);
+	if (res != GNOME_VFS_OK) {
+		return res;
+	}
+
+	old_name = gnome_vfs_unescape_string (old_uri->text, G_DIR_SEPARATOR_S);
+	new_name = gnome_vfs_unescape_string (new_uri->text, G_DIR_SEPARATOR_S);
+
+	if ((old_name == NULL) || (new_name == NULL)) {
+		g_free (old_name);
+		g_free (new_name);
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
+	quoted_old_name = g_shell_quote (old_name);
+	quoted_new_name = g_shell_quote (new_name);
+
+	cmd = g_strdup_printf ("mv %s %s", quoted_old_name, quoted_new_name);
+	res = ssh_connect (&handle, old_uri, cmd);
+	g_free (cmd);
+	g_free (old_name);
+	g_free (new_name);
+	g_free (quoted_old_name);
+	g_free (quoted_new_name);
+
+	if (res != GNOME_VFS_OK) {
+		return res;
+	}
+
+	ssh_wait_and_destroy (handle, context);
+
+	return res;
+}
+
 GnomeVFSResult
 do_unlink (GnomeVFSMethod *method,
 	   GnomeVFSURI *uri,
 	   GnomeVFSContext *context)
 {
 	SshHandle *handle = NULL;
-	char *cmd = NULL;
+	char *cmd;
 	GnomeVFSResult result;
 	gchar *name;
 	gchar *quoted_name;
 
 	name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
-	if (name == NULL)
+	if (name == NULL) {
 		return GNOME_VFS_ERROR_INVALID_URI;
+	}
 	quoted_name = g_shell_quote (name);
 
 	cmd = g_strdup_printf ("rm -rf %s", quoted_name);
@@ -818,7 +967,7 @@ do_unlink (GnomeVFSMethod *method,
 
 	return result;
 }
-#if 0
+
 static GnomeVFSResult
 do_check_same_fs (GnomeVFSMethod  *method,
 		GnomeVFSURI     *a,
@@ -853,7 +1002,6 @@ do_check_same_fs (GnomeVFSMethod  *method,
 
 	return GNOME_VFS_OK;
 }
-#endif
 
 static GnomeVFSResult
 do_set_file_info (GnomeVFSMethod *method,
@@ -868,8 +1016,9 @@ do_set_file_info (GnomeVFSMethod *method,
 	gchar *full_name;
 
 	full_name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
-	if (full_name == NULL)
+	if (full_name == NULL) {
 		return GNOME_VFS_ERROR_INVALID_URI;
+	}
 
 	if (mask & GNOME_VFS_SET_FILE_INFO_NAME) {
 		char *encoded_dir;
